/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <catch2/catch_test_macros.hpp>
#include <memory>
#include <vector>

#include "aidge/backend/cpu/operator/ConvDepthWiseImpl.hpp"
#include "aidge/backend/cpu/operator/ConvDepthWiseImpl_kernels.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/operator/ConvDepthWise.hpp"
#include "aidge/utils/TensorUtils.hpp"

using namespace Aidge;

TEST_CASE("[cpu/operator] ConvDepthWise(forward)", "[ConvDepthWise][CPU]") {
    SECTION("k[3,3]") {
        std::shared_ptr<Node> myCDW = ConvDepthWise(4, {3,3}, "mycdw");
        auto op = std::static_pointer_cast<OperatorTensor>(myCDW -> getOperator());
        std::shared_ptr<Tensor> myWeights = std::make_shared<Tensor>(Array4D<int,4,1,3,3> {
            {
                {{
                    {  0,  1,  2},
                    {  3,  4,  5},
                    {  6,  7,  8}

                }},
                {{
                    { 27, 28, 29},
                    { 30, 31, 32},
                    { 33, 34, 35}

                }},
                {{
                    { 54, 55, 56},
                    { 57, 58, 59},
                    { 60, 61, 62}
                }},
                {{
                    { 81, 82, 83},
                    { 84, 85, 86},
                    { 87, 88, 89}
                }}
            }
        });
        std::shared_ptr<Tensor> myBias = std::make_shared<Tensor>(Array1D<int,4> {{7,0,9,0}});
        std::shared_ptr<Tensor> myInput = std::make_shared<Tensor>(Array4D<int,2,4,5,5> { //NCHW
            {
                {
                    {{  0,   1,   2,   3,   4},
                    {  5,   6,   7,   8,   9},
                    { 10,  11,  12,  13,  14},
                    { 15,  16,  17,  18,  19},
                    { 20,  21,  22,  23,  24}},

                    {{ 25,  26,  27,  28,  29},
                    { 30,  31,  32,  33,  34},
                    { 35,  36,  37,  38,  39},
                    { 40,  41,  42,  43,  44},
                    { 45,  46,  47,  48,  49}},

                    {{ 50,  51,  52,  53,  54},
                    { 55,  56,  57,  58,  59},
                    { 60,  61,  62,  63,  64},
                    { 65,  66,  67,  68,  69},
                    { 70,  71,  72,  73,  74}},

                    {{ 75,  76,  77,  78,  79},
                    { 80,  81,  82,  83,  84},
                    { 85,  86,  87,  88,  89},
                    { 90,  91,  92,  93,  94},
                    { 95,  96,  97,  98,  99}}
                },
                {
                    {{100, 101, 102, 103, 104},
                    {105, 106, 107, 108, 109},
                    {110, 111, 112, 113, 114},
                    {115, 116, 117, 118, 119},
                    {120, 121, 122, 123, 124}},

                    {{125, 126, 127, 128, 129},
                    {130, 131, 132, 133, 134},
                    {135, 136, 137, 138, 139},
                    {140, 141, 142, 143, 144},
                    {145, 146, 147, 148, 149}},

                    {{150, 151, 152, 153, 154},
                    {155, 156, 157, 158, 159},
                    {160, 161, 162, 163, 164},
                    {165, 166, 167, 168, 169},
                    {170, 171, 172, 173, 174}},

                    {{175, 176, 177, 178, 179},
                    {180, 181, 182, 183, 184},
                    {185, 186, 187, 188, 189},
                    {190, 191, 192, 193, 194},
                    {195, 196, 197, 198, 199}}
                }
            }
        });
        std::shared_ptr<Tensor> myOutput = std::make_shared<Tensor>(Array4D<int,2,4,3,3> {
            {
                {
                    {{   319,    355,    391},
                    {   499,    535,    571},
                    {   679,    715,    751}},

                    {{  8745,   9024,   9303},
                    { 10140,  10419,  10698},
                    { 11535,  11814,  12093}},

                    {{ 29337,  29859,  30381},
                    { 31947,  32469,  32991},
                    { 34557,  35079,  35601}},

                    {{ 62061,  62826,  63591},
                    { 65886,  66651,  67416},
                    { 69711,  70476,  71241}}
                },
                {
                    {{  3919,   3955,   3991},
                    {  4099,   4135,   4171},
                    {  4279,   4315,   4351}},

                    {{ 36645,  36924,  37203},
                    { 38040,  38319,  38598},
                    { 39435,  39714,  39993}},

                    {{ 81537,  82059,  82581},
                    { 84147,  84669,  85191},
                    { 86757,  87279,  87801}},

                    {{138561, 139326, 140091},
                    {142386, 143151, 143916},
                    {146211, 146976, 147741}}
                }
            }
        });
        op -> associateInput(0, myInput);
        op -> associateInput(1, myWeights);
        op -> associateInput(2, myBias);
        op->setDataType(DataType::Int32);
        op->setBackend("cpu");
        myCDW -> forward();
        op -> getOutput(0) -> print();
        REQUIRE(*(op -> getOutput(0)) == *myOutput);
    }
    SECTION("point-wise") {
        ConvDepthWise_Op<2> conv_op = ConvDepthWise_Op<2>({1,1});
        std::shared_ptr<Tensor> weights = std::make_shared<Tensor>(std::vector<std::size_t>({3,1,1,1}));
        weights -> setBackend("cpu");
        std::shared_ptr<Tensor> biases = std::make_shared<Tensor>(std::vector<std::size_t>({3}));
        biases -> setBackend("cpu");
        std::shared_ptr<Tensor> input = std::make_shared<Tensor>(std::vector<std::size_t>({2,3,5,5}));
        input -> setBackend("cpu");
        std::shared_ptr<Tensor> expected_output = std::make_shared<Tensor>(std::vector<std::size_t>({2,3,5,5}));
        expected_output -> setBackend("cpu");

        float weighst_array[3] {-0.0045, -0.4223, -0.9452};
        weights->getImpl()->setRawPtr(weighst_array, 3);

        float biases_array[3] {-0.8595,  0.7062, -0.0062};
        biases->getImpl()->setRawPtr(biases_array, 3);

        float input_array[2*3*5*5] {
            0.6581, 0.2509, 0.2660, 0.8270, 0.8040, 0.3147, 0.5028, 0.2591, 0.8585,
            0.7762, 0.9972, 0.0305, 0.1202, 0.2682, 0.9306, 0.7927, 0.1494, 0.0678,
            0.5550, 0.4132, 0.4742, 0.6199, 0.1802, 0.6350, 0.2539, 0.5594, 0.0143,
            0.8656, 0.7105, 0.1420, 0.2464, 0.7883, 0.5715, 0.7642, 0.5492, 0.6628,
            0.4922, 0.7941, 0.8421, 0.7914, 0.0237, 0.8081, 0.0174, 0.6018, 0.7402,
            0.3770, 0.8786, 0.3651, 0.5355, 0.4267, 0.4457, 0.6756, 0.9631, 0.0145,
            0.4470, 0.5202, 0.2675, 0.5815, 0.3487, 0.3457, 0.7179, 0.0518, 0.1520,
            0.0573, 0.9219, 0.3615, 0.0866, 0.5237, 0.4725, 0.2565, 0.8726, 0.6434,
            0.6875, 0.2919, 0.3355, 0.1886, 0.1749, 0.0785, 0.4091, 0.1907, 0.4664,
            0.2738, 0.4784, 0.7807, 0.0687, 0.3091, 0.4557, 0.2277, 0.2424, 0.8691,
            0.1893, 0.2918, 0.5691, 0.1926, 0.2866, 0.0097, 0.5445, 0.5085, 0.1110,
            0.7099, 0.8927, 0.6182, 0.2538, 0.8694, 0.7872, 0.3196, 0.0710, 0.2888,
            0.0403, 0.1670, 0.6840, 0.7323, 0.4861, 0.3390, 0.1096, 0.5070, 0.3872,
            0.7473, 0.6224, 0.6910, 0.7530, 0.0149, 0.0866, 0.9022, 0.5027, 0.3849,
            0.5255, 0.1977, 0.0570, 0.9581, 0.5461, 0.4623, 0.0101, 0.2362, 0.5922,
            0.8398, 0.1497, 0.5160, 0.2862, 0.5931, 0.9728, 0.1353, 0.7790, 0.9137,
            0.9351, 0.4036, 0.7638, 0.3873, 0.0494, 0.7450};
        input->getImpl()->setRawPtr(input_array, 2*3*5*5);

        float expected_output_array[2*3*5*5] {
            -0.8624, -0.8606, -0.8607, -0.8632, -0.8631, -0.8609, -0.8617, -0.8606,
            -0.8633, -0.8629, -0.8639, -0.8596, -0.8600, -0.8607, -0.8636, -0.8630,
            -0.8601, -0.8598, -0.8620, -0.8613, -0.8616, -0.8622, -0.8603, -0.8623,
            -0.8606,  0.4700,  0.7002,  0.3407,  0.4062,  0.6463,  0.6022,  0.3733,
            0.4649,  0.3835,  0.4743,  0.4263,  0.4984,  0.3709,  0.3506,  0.3720,
            0.6962,  0.3650,  0.6989,  0.4521,  0.3936,  0.5470,  0.3352,  0.5520,
            0.4801,  0.5260, -0.4274, -0.6447, -0.9165, -0.0199, -0.4287, -0.4979,
            -0.2590, -0.5559, -0.3358, -0.3329, -0.6847, -0.0552, -0.1499, -0.0603,
            -0.8776, -0.3479, -0.0881, -0.5011, -0.4528, -0.2486, -0.8309, -0.6143,
            -0.6561, -0.2821, -0.3233, -0.8603, -0.8603, -0.8598, -0.8613, -0.8603,
            -0.8616, -0.8607, -0.8616, -0.8630, -0.8598, -0.8609, -0.8615, -0.8605,
            -0.8606, -0.8634, -0.8603, -0.8608, -0.8620, -0.8603, -0.8608, -0.8595,
            -0.8619, -0.8617, -0.8600, -0.8626,  0.3292,  0.4451,  0.5991,  0.3390,
            0.3738,  0.5712,  0.6762,  0.5843,  0.6892,  0.6357,  0.4174,  0.3969,
            0.5009,  0.5631,  0.6599,  0.4921,  0.5427,  0.3906,  0.4434,  0.4144,
            0.3882,  0.6999,  0.6697,  0.3252,  0.4939, -0.3700, -0.5029, -0.1931,
            -0.0601, -0.9118, -0.5224, -0.4432, -0.0157, -0.2294, -0.5660, -0.7999,
            -0.1477, -0.4939, -0.2767, -0.5668, -0.9257, -0.1341, -0.7425, -0.8698,
            -0.8900, -0.3877, -0.7282, -0.3722, -0.0529, -0.7103};
        expected_output->getImpl()->setRawPtr(expected_output_array, 2*3*5*5);

        conv_op.associateInput(0, input);
        conv_op.associateInput(1, weights);
        conv_op.associateInput(2, biases);

        conv_op.setBackend("cpu");
        conv_op.setDataType(DataType::Float32);
        conv_op.forwardDims();

        conv_op.forward();

        conv_op.getOutput(0)->print();

        REQUIRE(approxEq<float>(*(conv_op.getOutput(0)),*expected_output, 1e-3f, 1e-4f));
    }
}